#!/usr/bin/env node
'use strict';

const fs = require('fs');
const https = require('https');
const path = require('path');
const exec = require('child_process').exec;
const pkgJson = require(path.resolve(__dirname + '/../package.json'));
const moduleName = pkgJson.binary.module_name;
const version = pkgJson.version;

/**
 * List of supported Node.js module API versions.
 */
const nodeModuleVersions = Object.values(pkgJson.binary.targets);
const bindingDir = path.dirname(path.resolve(__dirname + '/../' + pkgJson.binary.module_path));
const prefix = pkgJson.binary.host + '/' + pkgJson.binary.remote_path
	.replace(/^.\//, '')
	.replace(/{name}/, pkgJson.name)
	.replace(/{version}/, version) + '/';
let failed = false;

if (!fs.existsSync(bindingDir)) {
	fs.mkdirSync(bindingDir)
}

function download(abiVersion) {
	return new Promise((resolve, reject) => {
		var filename = `${moduleName}-v${version}-node-v${abiVersion}-${process.platform}-${process.arch}.tar.gz`;
		var url = prefix + filename;
		var dest = path.join(bindingDir, filename);

		console.log(`Downloading ${url}`);

		function handleResponse(response) {
			if (response.statusCode === 302 || response.statusCode === 302) {
				return https.get(response.headers.location, handleResponse);
			}

			if (response.statusCode !== 200) {
				console.log(`Failed to download ${url} (${response.statusCode})`);
				failed = true;
				resolve();
				return;
			}

			console.log(`Downloaded ${url} successfully (${response.statusCode})`);
			response.pipe(fs.createWriteStream(dest))
				.on('finish', function () {
					console.log(`Extracting ${filename}...`);
					exec(`tar xzf "${filename}"`, { cwd: bindingDir }, function (error) {
						if (error) {
							console.log(`Failed to extract ${filename}: ${error.message || error.toString()}`);
							failed = true;
							reject(error);
						} else {
							fs.unlinkSync(dest);
							resolve();
						}
					});
				});
		}
		https.get(url, handleResponse).on('error', reject);
	});
}

Promise
	.all(nodeModuleVersions.map(download))
	.then(() => {
		if (failed) {
			console.log('Done, but with errors');
			process.exit(1);
		}
		console.log('Done!');
	})
	.catch(err => {
		console.log(err);
		process.exit(1);
	});
